# -*- coding: ISO-8859-1 -*-
# Copyright (C) 2003-2005 Juan David Ibez Palomar <jdavid@itaapy.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

# Import from itools
from itools import types
from itools.handlers import Text
from itools.xml import namespaces, parser, XML
from itools.zope import get_context


#############################################################################
# Namespace
#############################################################################

class Record(object):

    @classmethod
    def to_unicode(cls, value):
        lines = []
        for key, value in value.items():
            prefix, local_name = key
            namespace = namespaces.get_namespace_by_prefix(prefix)
            schema = namespace.get_attribute_schema(local_name)
            type = schema['type']
            value = type.to_unicode(value)
            qname = types.QName.encode(key)
            lines.append('\n    <%s>%s</%s>' % (qname, value, qname))
        return ''.join(lines) + '\n'



class Namespace(namespaces.AbstractNamespace):

    class_uri = 'http://xml.ikaaro.org/namespaces/metadata'
    class_prefix = 'ikaaro'


    @classmethod
    def get_attribute_schema(cls, local_name):
        try:
            return cls.schema[local_name]
        except KeyError:
            raise XML.XMLError, 'unknown property "%s"' % local_name


    schema = {
##        'format': {'type': types.String, 'default': None},
##        'version': {'type': types.String, 'default': None},
##        'owner': {'type': types.String, 'default': None},
        # LocaleAware
        'language': {'type': types.String, 'default': None},
        'isVersionOf': {'type': types.String, 'default': None},
        'hasVersion': {'type': types.String, 'default': None},
        # Workflow
        'state': {'type': types.String, 'default': 'private'},
        'wf_transition': {'type': Record, 'default': []},
##        'name': {'type': types.String, 'default': None},
##        'user': {'type': types.String, 'default': None},
##        'comments': {'type': types.Unicode, 'default': u''},
        # Archive
##        'id': {'type': types.String, 'default': None},
        # Users
        'user_theme': {'type': types.String, 'default': 'surf'},
        'user_language': {'type': types.String, 'default': 'en'},
        'website_is_open': {'type': types.Boolean, 'default': False},
        'website_languages': {'type': types.Tokens, 'default': ('en',)}}


namespaces.set_namespace(Namespace)



#############################################################################
# Handler
#############################################################################
class Metadata(Text.Text):

    def get_skeleton(self, handler_class=None, **kw):
        # Initialize the properties to add
        properties = {'format': handler_class.class_id,
                      'version': handler_class.class_version}
        # Update properties with the keyword parameters
        properties.update(kw)

        needed_namespaces = {}
        # Build the XML
        skeleton = ['<?xml version="1.0" encoding="UTF-8"?>']
        for name, value in properties.items():
            if value is not None:
                prefix, local_name = types.QName.decode(name)
                if prefix is None:
                    type = types.String
                else:
                    namespace = namespaces.get_namespace_by_prefix(prefix)
                    needed_namespaces[namespace.class_uri] = prefix
                    schema = namespace.get_attribute_schema(local_name)
                    type = schema['type']

                if isinstance(value, dict):
                    for lang, value in value.items():
                        value = type.encode(value)
                        skeleton.append('  <%s xml:lang="%s">%s</%s>'
                                        % (name, lang, value, name))
                else:
                    value = type.encode(value)
                    skeleton.append('  <%s>%s</%s>' % (name, value, name))

        # Insert open root element with the required namespace declarations
        if needed_namespaces:
            needed_namespaces = [ 'xmlns:%s="%s"' % (y, x)
                                  for x, y in needed_namespaces.items() ]
            head = '<metadata %s>' % '\n          '.join(needed_namespaces)
        else:
            head = '<metadata>'
        skeleton.insert(1, head)

        skeleton.append('</metadata>')
        return '\n'.join(skeleton)


    def _load_state(self, resource):
        state = self.state
        # Keep the namespace prefixes
        state.prefixes = set()

        p_key = None
        p_type = None
        p_language = None
        p_value = ''
        stack = []
        for event, value, line_number in parser.parse(resource.read()):
            if event == parser.START_ELEMENT:
                namespace_uri, prefix, local_name = value
                if local_name == 'metadata':
                    stack.append({})
                else:
                    # Update prefixes
                    if prefix is not None:
                        state.prefixes.add(prefix)
                    # Build the property key
                    p_key = (prefix, local_name)
                    # Get the property type
                    namespace = namespaces.get_namespace(namespace_uri)
                    schema = namespace.get_attribute_schema(local_name)
                    p_type = schema['type']

                    # XXX Backwards compatibility (to be removed by 0.14)
                    if namespace_uri is None and local_name == 'wf_transition':
                        p_type = Record

                    if p_type is Record:
                        stack.append({})
                    else:
                        p_value = ''
            elif event == parser.END_ELEMENT:
                namespace, prefix, local_name = value
                # Build the property key
                p_key = (prefix, local_name)
                # Get the property type
                namespace = namespaces.get_namespace(namespace)
                schema = namespace.get_attribute_schema(local_name)
                p_type = schema['type']
                p_default = schema.get('default')

                # XXX Backwards compatibility (to be removed by 0.14)
                if namespace_uri is None and local_name == 'wf_transition':
                    p_type = Record
                    p_default = []

                if local_name == 'metadata':
                    state.properties = stack.pop()
                else:
                    # Decode value
                    if p_type is Record:
                        p_value = stack.pop()
                    elif p_type is types.Unicode:
                        p_value = p_type.decode(p_value, 'UTF-8')
                    else:
                        p_value = p_type.decode(p_value)
                    # Set property
                    if isinstance(p_default, list):
                        stack[-1].setdefault(p_key, []).append(p_value)
                    elif p_language is None:
                        stack[-1][p_key] = p_value
                    else:
                        stack[-1].setdefault(p_key, {})
                        stack[-1][p_key][p_language] = p_value
                    # Reset variables
                    p_type = None
                    p_language = None
                    p_value = ''
            elif event == parser.TEXT:
                if p_key is not None:
                    p_value += value
            elif event == parser.ATTRIBUTE:
                namespace_uri, prefix, local_name, value = value
                if (namespace_uri == namespaces.XMLNamespace.class_uri
                    and local_name == 'lang'):
                    p_language = value
                # XXX Backwards compatibility (to be removed by 0.14)
                elif namespace_uri is None and local_name == 'lang':
                    p_language = value


    def to_unicode(self, encoding='UTF-8'):
        state = self.state

        lines = [u'<?xml version="1.0" encoding="%s"?>' % encoding]

        # Insert open root element with the required namespace declarations
        if state.prefixes:
            aux = [ (x, namespaces.get_namespace_by_prefix(x).class_uri)
                    for x in state.prefixes ]
            aux = u'\n          '.join([ u'xmlns:%s="%s"' % x for x in aux ])
            lines.append(u'<metadata %s>' % aux)
        else:
            lines.append(u'<metadata>')

        for key, value in state.properties.items():
            prefix, local_name = key

            # Get the type
            namespace = namespaces.get_namespace_by_prefix(prefix)
            schema = namespace.get_attribute_schema(local_name)
            type = schema['type']
            # Get the qualified name
            if prefix is None:
                qname = local_name
            else:
                qname = '%s:%s' % key

            if isinstance(value, dict):
                for language, value in value.items():
                    value = type.to_unicode(value)
                    lines.append(u'  <%s xml:lang="%s">%s</%s>' \
                                 % (qname, language, value, qname))
            elif isinstance(value, list):
                for value in value:
                    value = type.to_unicode(value)
                    lines.append(u'  <%s>%s</%s>' % (qname, value, qname))
            else:
                value = type.to_unicode(value)
                lines.append(u'  <%s>%s</%s>' % (qname, value, qname))

        lines.append(u'</metadata>')
        return u'\n'.join(lines)


    ########################################################################
    # API
    ########################################################################
    def get_property(self, name, language=None):
        key = types.QName.decode(name)

        # Default value
        prefix, local_name = key
        namespace = namespaces.get_namespace_by_prefix(prefix)
        schema = namespace.get_attribute_schema(local_name)
        default_value = schema.get('default', None)

        state = self.state
        if key in state.properties:
            value = state.properties[key]
        else:
            return default_value

        if isinstance(value, dict):
            # LocaleAware
            if language is None:
                # Language negotiation
                accept = get_context().request.accept_language
                available_languages = [ k for k, v in value.items()
                                        if v.strip() ]
                language = accept.select_language(available_languages)
                if language is None:
                    # Default (XXX pick one at random XXX)
                    language = value.keys()[0]
                return value[language]
            return value.get(language, default_value)
        return value


    def has_property(self, name, language=None):
        key = types.QName.decode(name)

        state = self.state
        if key not in state.properties:
            return False

        if language is not None:
            return language in state.properties[key]

        return True


    def set_property(self, name, value, language=None):
        self.set_changed()

        key = types.QName.decode(name)

        state = self.state
        # Set the value
        if language is None:
            prefix, local_name = key
            namespace = namespaces.get_namespace_by_prefix(prefix)
            schema = namespace.get_attribute_schema(local_name)

            type = schema['type']
            default = schema.get('default', None)
            if isinstance(default, list):
                if isinstance(value, list):
                    state.properties[key] = value
                else:
                    values = state.properties.setdefault(key, [])
                    values.append(value)
            else:
                state.properties[key] = value
        else:
            values = state.properties.setdefault(key, {})
            values[language] = value

        # Update prefixes
        if key[0] is not None:
            state.prefixes.add(key[0])
        if isinstance(value, dict):
            for prefix, local_name in value:
                if prefix is not None:
                    state.prefixes.add(prefix)


    def del_property(self, name, language=None):
        key = types.QName.decode(name)

        state = self.state
        if key in state.properties:
            if language is None:
                self.set_changed()
                del state.properties[key]
            else:
                values = state.properties[key]
                if language in values:
                    self.set_changed()
                    del values[language]
